;;;============================================================================

;;; File: "gsi.scm"

;;; Copyright (c) 2013 by Marc Feeley, All Rights Reserved.

(##include "~~lib/gambit#.scm")


(define (##main-gsi/gsc)
  (##repl-debug-main))

(##namespace (""))

;;;============================================================================

(##c-declare #<<end-of-c-declare

#include <stdio.h>
#include <emscripten/emscripten.h>

#define SCHEME_LIBRARY_LINKER ____20_gsi__

___BEGIN_C_LINKAGE
extern ___mod_or_lnk SCHEME_LIBRARY_LINKER (___global_state_struct*);
___END_C_LINKAGE

em_callback_func mycallback()
{
}

void main()
{

 EM_ASM(

    function web_trampoline() {
        _heartbeat_interrupt();
        var wait = _idle();
        if (wait < 0) {
            _cleanup();
        } else {
            console.log('wait=' + wait);
            setTimeout(web_trampoline, Math.max(1, Math.round(1000*wait)));
        }
    };

    _setup();

    web_trampoline();

    );

  emscripten_set_main_loop(mycallback, 10, 1);
}

extern void setup() {

  /*
   * Setup the Scheme library by calling "___setup" with appropriate
   * parameters.  The call to "___setup_params_reset" sets all parameters
   * to their default setting.
   */

  ___setup_params_struct setup_params;

  ___setup_params_reset(&setup_params);

  setup_params.version = ___VERSION;
  setup_params.linker  = SCHEME_LIBRARY_LINKER;

  ___ON_THROW(___setup(&setup_params);,);
}

extern void cleanup() {

  /* Cleanup the Scheme library */

  ___cleanup();
}

extern void user_interrupt() {
  ___raise_interrupt(___INTR_USER);
}

extern void heartbeat_interrupt() {
  ___raise_interrupt(___INTR_HEARTBEAT);
}

extern ___SCMOBJ scheme_idle() {
#define ___NARGS 0
  ___BEGIN_SFUN_SCMOBJ(___GLO__23__23_idle)
  ___BEGIN_SFUN_BODY
  ___SFUN_CALL_SCMOBJ
  ___SFUN_SET_RESULT_SCMOBJ
  ___END_SFUN_BODY
  ___SFUN_ERROR_SCMOBJ
  ___SFUN_SET_RESULT_SCMOBJ
  ___END_SFUN_SCMOBJ
  return ___result;
#undef ___NARGS
}

extern double idle() {

  ___processor_state ___ps = ___PSTATE;
  ___SCMOBJ result = ___FIX(0);

  ___ON_THROW(result = scheme_idle();,);

  if (___FIXNUMP(result))
    return -1.0; /* signal program termination */
  else
    return ___FLONUM_VAL(result);
}


end-of-c-declare
)

(define (##idle)

  ;; interrupts must be checked here to avoid overflowing the stack
  ;; when there is continuous context-switching to threads which are
  ;; idle (not checking stack overflows)

  ;; return interval until next heartbeat
  (##next-heartbeat-interval))

(define (##next-heartbeat-interval)

  (##declare (not interrupts-enabled))

  (let* ((run-queue (macro-run-queue))
         (runnable-threads?
          (##not
           (let ((root (macro-btq-left run-queue)))
             (and (##not (##eq? root run-queue))
                  (##eq? (macro-btq-left root) run-queue)
                  (##eq? (macro-btq-right root) run-queue))))))

    (if runnable-threads?

        (begin
          ;; There are other threads that can run, so request
          ;; to call "heartbeat" real soon to run those threads.
          interval-runnable)

        (let* ((next-sleeper
                (macro-toq-leftmost run-queue))
               (sleep-interval
                (if (##eq? next-sleeper run-queue)
                    +inf.0
                    (begin
                      ;; There is a sleeping thread, so figure out in
                      ;; how much time it needs to wake up.
                      (##flmax
                       (##fl- (macro-thread-timeout next-sleeper)
                              (##current-time-point))
                       interval-min-wait))))
               (next-condvar
                (macro-btq-deq-next run-queue))
               (io-interval
                (if (##eq? next-condvar run-queue)
                    interval-no-io-pending ;; I/O is not pending, just relax
                    interval-io-pending))) ;; I/O is pending, so come back soon
          (##flmin sleep-interval io-interval)))))

(define interval-runnable 0.0)
(set! interval-runnable 0.0)

(define interval-io-pending 0.0)
(set! interval-io-pending 0.001)

(define interval-no-io-pending 0.0)
(set! interval-no-io-pending 1.0)

(define interval-min-wait 0.0)
(set! interval-min-wait 0.0001)

;;;----------------------------------------------------------------------------

(define ##jseval
  (c-lambda (char-string) char-string "emscripten_run_script_string"))

(define jseval ##jseval)

(define (##show-definition-of subject)
  (let ((s
         (cond ((##procedure? subject)
                (##object->string (##procedure-name subject)))
               (else
                (##object->string subject)))))
    (##jseval
     (##string-append "open(\"http://www.iro.umontreal.ca/~gambit/doc/gambit-c.html#"
                      (##escape-link (##string-append "Definition of " s))
                      "\")")))
    (##void))

(set! ##help-hook ##show-definition-of)

;;;----------------------------------------------------------------------------

(define (term-esc cmd) ;; Terminus escape sequence
  (let ((port (repl-output-port)))
    (tty-mode-set! port #t #t #t #t) ;; Gambit's tty driver needs to pass the escapes as-is
    (print port: port "\x1B[?0y" cmd)
    (tty-mode-set! port #t #t #f #f)))

;;;----------------------------------------------------------------------------

;; use line buffering for REPL output

;;(port-settings-set! (repl-output-port) '(output-buffering: line))

;;;----------------------------------------------------------------------------

;; Remember REPL history from one run of Gambit to the next.

#;
(define (##setup-repl-history)
  (let ((input-port (repl-input-port)))
    (if (##tty? input-port)
        (let ((history
               (##jseval "var history;
                          history = sessionStorage.getItem('gambit-repl-history');
                          if (typeof history !== 'string') history = '';
                          history;")))

          (if (##string? history)
              (##tty-history-set! input-port history))

          (##add-exit-job!
           (lambda ()
             (let ((history (##tty-history input-port)))

               (##jseval (##string-append "sessionStorage.setItem('gambit-repl-history',"
                                          (##object->string history)
                                          "); '';")))))))))

;;;----------------------------------------------------------------------------

;; Start the main REPL in the primordial thread, and create a service
;; thread which executes the rest of the program (returning back from
;; the C call to ___setup).

(define (run-service-thread denv)

  ;; disable user interrupts in service thread so that
  ;; the idle function won't start a REPL of its own.
  (##thread-save!
   (lambda (thread)
     (macro-denv-interrupt-mask-set! denv 1)
     (macro-thread-denv-set! thread denv)
     (##thread-restore! thread (lambda () #f)))))

(define (run-primordial-thread)
  (##main-gsi/gsc))

(##main-set!
 (lambda ()
   (continuation-capture
    (lambda (cont)

      (thread-start!
       (make-root-thread
        (lambda ()
          (let ((denv (macro-thread-denv (current-thread))))
            (continuation-graft cont run-service-thread denv)))))

      (continuation-graft cont run-primordial-thread)))))

;;;============================================================================

